package main

import (
	"encoding/json"
	"flag"
	"fmt"
	"hash/crc32"
	"log"
	"net/url"
	"os"
	"strconv"
	"strings"

	"git.autistici.org/ale/autoradio"
	"github.com/gonuts/commander"
	gonutsflag "github.com/gonuts/flag"
)

// Format for output of structured data.
var outputFormat = flag.String("format", "txt", "Output format for structured data (json, txt)")

type optionalValue struct {
	isSet bool
}

func (v *optionalValue) IsSet() bool {
	return v.isSet
}

type stringOptionalValue struct {
	*optionalValue
	value string
}

func newStringOptionalValue() *stringOptionalValue {
	return &stringOptionalValue{}
}

func (v *stringOptionalValue) Set(val string) error {
	v.isSet = true
	v.value = val
	return nil
}

func (v *stringOptionalValue) Value() string {
	return v.value
}

func (v *stringOptionalValue) String() string {
	return v.value
}

func (v *stringOptionalValue) Get() interface{} {
	return v.value
}

type intOptionalValue struct {
	*optionalValue
	value int
}

func newIntOptionalValue() *intOptionalValue {
	return &intOptionalValue{}
}

func (v *intOptionalValue) Set(val string) error {
	ival, err := strconv.Atoi(val)
	v.value = ival
	v.isSet = true
	return err
}

func (v *intOptionalValue) Value() int {
	return v.value
}

func (v *intOptionalValue) String() string {
	return fmt.Sprintf("%v", v.value)
}

func (v *intOptionalValue) Get() interface{} {
	return v.value
}

type floatOptionalValue struct {
	*optionalValue
	value float64
}

func newFloatOptionalValue() *floatOptionalValue {
	return &floatOptionalValue{}
}

func (v *floatOptionalValue) Set(val string) error {
	fval, err := strconv.ParseFloat(val, 64)
	v.value = fval
	v.isSet = true
	return err
}

func (v *floatOptionalValue) Value() float64 {
	return v.value
}

func (v *floatOptionalValue) String() string {
	return fmt.Sprintf("%v", v.value)
}

func (v *floatOptionalValue) Get() interface{} {
	return v.value
}

type HasAddFlags interface {
	AddFlags(*gonutsflag.FlagSet)
}

type CommandInterface interface {
	Run([]string)
	Command() *commander.Command
}

type BaseCommand struct {
	UsageLine string
	Short     string
	Long      string
}

func (b *BaseCommand) Command() *commander.Command {
	return &commander.Command{
		UsageLine: b.UsageLine,
		Short:     b.Short,
		Long:      b.Long,
	}
}

func (b *BaseCommand) Run(args []string) {
}

func getClient() *autoradio.Client {
	return autoradio.NewClient(autoradio.NewEtcdClient(false))
}

func generateUsername(path string) string {
	return fmt.Sprintf("source%d", crc32.ChecksumIEEE([]byte(path)))
}

func setRelay(m *autoradio.Mount, relayUrl string) {
	if relayUrl == "" {
		// Randomly generate source credentials.
		m.Username = generateUsername(m.Name)
		m.Password = autoradio.GeneratePassword()
	} else {
		// Validate the given relay URL.
		u, err := url.Parse(relayUrl)
		if err != nil {
			log.Fatal(err)
		}
		m.RelayUrl = u.String()
		m.Username = ""
		m.Password = ""
	}
}

func printMount(m *autoradio.Mount) {
	switch *outputFormat {
	case "json":
		s, _ := json.MarshalIndent(m, "", "    ")
		os.Stdout.Write(s)

	case "txt":
		fmt.Printf("path=%s\n", m.Name)
		if m.Username != "" {
			fmt.Printf("username=%s\npassword=%s\n", m.Username, m.Password)
		}
		if m.RelayUrl != "" {
			fmt.Printf("relay_url=%s\n", m.RelayUrl)
		}
		if m.Fallback != "" {
			fmt.Printf("fallback_url=%s\n", m.Fallback)
		}
		if t := m.Transcoding; t != nil {
			fmt.Printf("transcode_source_url=%s\n", t.SourceName)
			fmt.Printf("transcode_format=%s\n", t.Format)
			fmt.Printf("transcode_bitrate=%d\n", t.BitRate)
			fmt.Printf("transcode_quality=%f\n", t.Quality)
			fmt.Printf("transcode_samplerate=%d\n", t.SampleRate)
		}

	default:
		log.Printf("unsupported output format \"%s\"", *outputFormat)
	}
}

func mountExists(name string, client *autoradio.Client) bool {
	m, _ := client.GetMount(name)
	return m != nil
}

// Create a new mountpoint.
type createMountCommand struct {
	BaseCommand
	relay    string
	fallback string
}

func newCreateMountCommand() *createMountCommand {
	return &createMountCommand{
		BaseCommand: BaseCommand{
			UsageLine: "create-mount <path>",
			Short:     "Create a new mountpoint",
			Long: `
Create a new mountpoint at the specified PATH (which will become the
absolute URL of the stream).
`,
		},
	}
}

func (cmd *createMountCommand) AddFlags(f *gonutsflag.FlagSet) {
	f.StringVar(&cmd.relay, "relay", "", "Upstream URL to relay")
	f.StringVar(&cmd.fallback, "fallback", "", "Fallback stream URL")
}

func (cmd *createMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}
	fmt.Printf("args: %v\n", args)

	path := args[0]
	if !strings.HasPrefix(path, "/") {
		log.Printf("Warning: mountpoint %s does not start with a slash, using /%s instead", path, path)
		path = "/" + path
	}

	// Check if the mount already exists.
	client := getClient()
	if mountExists(path, client) {
		log.Fatal("ERROR: A mount with that name already exists!")
	}

	// Create the new mount and set the relevant fields (depending
	// on the options passed to the command).
	m := &autoradio.Mount{Name: path}
	setRelay(m, cmd.relay)
	m.Fallback = cmd.fallback

	if err := m.Valid(); err != nil {
		log.Fatalf("ERROR: mount configuration is invalid: %v", err)
	}

	if err := client.SetMount(m); err != nil {
		log.Fatalf("ERROR: creating mount: %v", err)
	}

	printMount(m)
}

// Create a submount (transcoded stream).
type createTranscodingMountCommand struct {
	BaseCommand
	params   *autoradio.EncodingParams
	fallback string
}

func newCreateTranscodingMountCommand() *createTranscodingMountCommand {
	return &createTranscodingMountCommand{
		BaseCommand: BaseCommand{
			UsageLine: "create-transcoding-mount <path>",
			Short:     "Create a transcoding mount",
			Long: `
Create a new stream that will transcode the parent stream with
different encoding parameters.
`,
		},
		params: autoradio.NewEncodingParams(),
	}
}

func addEncodingFlags(f *gonutsflag.FlagSet, p *autoradio.EncodingParams) {
	f.StringVar(&p.SourceName, "source", "", "Source mountpoint")
	f.StringVar(&p.Format, "codec", p.Format, "Encoding format")
	f.Float64Var(&p.Quality, "quality", p.Quality, "Quality (for VBR encoders)")
	f.IntVar(&p.BitRate, "bitrate", p.BitRate, "Bitrate (kbps)")
	f.IntVar(&p.SampleRate, "samplerate", p.SampleRate, "Sample rate (Hz)")
	f.IntVar(&p.Channels, "channels", p.Channels, "Number of channels")
	f.StringVar(&p.StereoMode, "stereo-mode", p.StereoMode, "Stereo mode for mp3 codec (stereo, joint_stereo)")
}

func (cmd *createTranscodingMountCommand) AddFlags(f *gonutsflag.FlagSet) {
	addEncodingFlags(f, cmd.params)
	f.StringVar(&cmd.fallback, "fallback", "", "Fallback stream URL")
}

func (cmd *createTranscodingMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}

	path := args[0]
	if !strings.HasPrefix(path, "/") {
		log.Printf("Warning: mountpoint %s does not start with a slash, using /%s instead", path, path)
		path = "/" + path
	}

	// The mount path should not exist.
	client := getClient()
	if mountExists(path, client) {
		log.Fatal("ERROR: a mount with that name already exists!")
	}
	// The source mount should exist.
	if !mountExists(cmd.params.SourceName, client) {
		log.Fatal("ERROR: the source mount does not exist!")
	}

	// Retrieve the parent mount point and add a TranscodingMount.
	m := &autoradio.Mount{
		Name:        path,
		Transcoding: cmd.params,
	}
	setRelay(m, "")
	m.Fallback = cmd.fallback

	if err := m.Valid(); err != nil {
		log.Fatalf("ERROR: mount configuration is invalid: %v", err)
	}

	if err := client.SetMount(m); err != nil {
		log.Fatalf("ERROR: creating mount: %v", err)
	}

	printMount(m)
}

// Edit a mountpoint.
type editMountCommand struct {
	BaseCommand
	relay           *stringOptionalValue
	fallback        *stringOptionalValue
	transFormat     *stringOptionalValue
	transSource     *stringOptionalValue
	transBitRate    *intOptionalValue
	transSampleRate *intOptionalValue
	transChannels   *intOptionalValue
	transStereoMode *stringOptionalValue
	transQuality    *floatOptionalValue
}

var UNSET = "UNSET"

func newEditMountCommand() *editMountCommand {
	return &editMountCommand{
		BaseCommand: BaseCommand{
			UsageLine: "edit-mount <path>",
			Short:     "Edit an existing mountpoint",
			Long: `
Modify parameters of the specified mountpoint, such as the relay
and the fallback URL. If the relay option is set, the mountpoint
will not accept source connections anymore. To revert to the
default, non-relay behavior, set the relay to the empty string
(with --relay="").
`,
		},
		relay:           newStringOptionalValue(),
		fallback:        newStringOptionalValue(),
		transFormat:     newStringOptionalValue(),
		transSource:     newStringOptionalValue(),
		transBitRate:    newIntOptionalValue(),
		transSampleRate: newIntOptionalValue(),
		transChannels:   newIntOptionalValue(),
		transStereoMode: newStringOptionalValue(),
		transQuality:    newFloatOptionalValue(),
	}
}

func (cmd *editMountCommand) AddFlags(f *gonutsflag.FlagSet) {
	f.Var(cmd.relay, "relay", "Upstream URL to relay")
	f.Var(cmd.fallback, "fallback", "Fallback stream URL")

	f.Var(cmd.transSource, "source", "[transcoding] Source mountpoint")
	f.Var(cmd.transFormat, "codec", "[transcoding] Encoding format")
	f.Var(cmd.transQuality, "quality", "[transcoding] Quality (for VBR encoders)")
	f.Var(cmd.transBitRate, "bitrate", "[transcoding] Bitrate (kbps)")
	f.Var(cmd.transSampleRate, "samplerate", "[transcoding] Sample rate (Hz)")
	f.Var(cmd.transChannels, "channels", "[transcoding] Number of channels")
	f.Var(cmd.transStereoMode, "stereo-mode", "[transcoding] Stereo mode for mp3 encoding (stereo, joint_stereo)")
}

func (cmd *editMountCommand) transcodingOptionsSet() bool {
	opts := []interface {
		IsSet() bool
	}{
		cmd.transSource, cmd.transFormat, cmd.transQuality,
		cmd.transBitRate, cmd.transSampleRate, cmd.transChannels,
		cmd.transStereoMode,
	}
	for _, o := range opts {
		if o.IsSet() {
			return true
		}
	}
	return false
}

func (cmd *editMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}

	client := getClient()
	m, err := client.GetMount(args[0])
	if err != nil {
		log.Fatalf("ERROR: %v", err)
	}
	if m == nil {
		log.Fatal("ERROR: mount not found")
	}

	// Only set those fields that were passed on the command line.
	if cmd.fallback.IsSet() {
		m.Fallback = cmd.fallback.Value()
	}
	if cmd.relay.IsSet() {
		setRelay(m, cmd.relay.Value())
	}

	if cmd.transcodingOptionsSet() && m.Transcoding == nil {
		log.Fatal("ERROR: can't set transcoding options on a non-transcoding mount (delete and re-create)")
	}

	if cmd.transFormat.IsSet() {
		m.Transcoding.Format = cmd.transFormat.Value()
	}
	if cmd.transSource.IsSet() {
		m.Transcoding.SourceName = cmd.transSource.Value()
	}
	if cmd.transBitRate.IsSet() {
		m.Transcoding.BitRate = cmd.transBitRate.Value()
	}
	if cmd.transSampleRate.IsSet() {
		m.Transcoding.SampleRate = cmd.transSampleRate.Value()
	}
	if cmd.transQuality.IsSet() {
		m.Transcoding.Quality = cmd.transQuality.Value()
	}
	if cmd.transChannels.IsSet() {
		m.Transcoding.Channels = cmd.transChannels.Value()
	}
	if cmd.transStereoMode.IsSet() {
		m.Transcoding.StereoMode = cmd.transStereoMode.Value()
	}

	if err := m.Valid(); err != nil {
		log.Fatalf("ERROR: mount configuration is invalid: %v", err)
	}

	if err := client.SetMount(m); err != nil {
		log.Fatalf("ERROR: updating mount: %v", err)
	}

	printMount(m)
}

// Delete an existing mountpoint.
type deleteMountCommand struct {
	BaseCommand
}

func newDeleteMountCommand() *deleteMountCommand {
	return &deleteMountCommand{
		BaseCommand{
			UsageLine: "delete-mount <path>",
			Short:     "Delete a mountpoint",
			Long: `
Delete the specified mountpoint.
`,
		},
	}
}

func (cmd *deleteMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}
	path := args[0]
	client := getClient()
	if !mountExists(path, client) {
		log.Fatal("ERROR: mount not found")
	}

	if err := client.DelMount(path); err != nil {
		log.Fatalf("ERROR: deleting mount: %v", err)
	}

	// Delete all the transcoding mounts that have this as a
	// source.
	mounts, err := client.ListMounts()
	if err != nil {
		log.Fatalf("ERROR: %v", err)
	}
	for _, m := range mounts {
		if m.HasTranscoder() && m.Transcoding.SourceName == path {
			if err := client.DelMount(m.Name); err != nil {
				log.Printf("ERROR: deleting transcoded mount %s: %v", m.Name, err)
			}
		}
	}

	log.Printf("mountpoint %s removed", path)
}

// List known mountpoints.
type listMountsCommand struct {
	BaseCommand
}

func newListMountsCommand() *listMountsCommand {
	return &listMountsCommand{
		BaseCommand{
			UsageLine: "list-mounts",
			Short:     "List all configured mountpoints",
			Long: `
Outputs a list of all the currently configured mountpoints.
`,
		},
	}
}

func (cmd *listMountsCommand) Run(args []string) {
	if len(args) != 0 {
		log.Fatal("Too many arguments")
	}

	mounts, err := getClient().ListMounts()
	if err != nil {
		log.Fatalf("ERROR: %v", err)
	}
	var names []string
	for _, m := range mounts {
		names = append(names, m.Name)
	}

	switch *outputFormat {
	case "json":
		s, _ := json.MarshalIndent(names, "", "    ")
		os.Stdout.Write(s)

	//case "txt":
	default:
		for _, n := range names {
			fmt.Println(n)
		}
	}
}

// Show mountpoint information.
type showMountCommand struct {
	BaseCommand
}

func newShowMountCommand() *showMountCommand {
	return &showMountCommand{
		BaseCommand{
			UsageLine: "show-mount <path>",
			Short:     "Show mountpoint information",
			Long: `
Print information about the specified mountpoint (including
the source credentials).
`,
		},
	}
}

func (cmd *showMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}

	m, err := getClient().GetMount(args[0])
	if err != nil {
		log.Fatalf("ERROR: %v", err)
	}
	if m == nil {
		log.Fatal("ERROR: mount not found")
	}

	printMount(m)
}

var cmdr = &commander.Command{
	UsageLine: "radioctl <command> [<args>...]",
	Short:     "Manage `autoradio' configuration",
}

func addCommand(c CommandInterface) {
	cmd := c.Command()
	cmd.Run = func(unused *commander.Command, args []string) error {
		c.Run(args)
		return nil
	}
	if afc, ok := c.(HasAddFlags); ok {
		afc.AddFlags(&cmd.Flag)
	}
	cmdr.Subcommands = append(cmdr.Subcommands, cmd)
}

func init() {
	addCommand(newCreateMountCommand())
	addCommand(newCreateTranscodingMountCommand())
	addCommand(newEditMountCommand())
	addCommand(newDeleteMountCommand())
	addCommand(newListMountsCommand())
	addCommand(newShowMountCommand())
	flag.Usage = usage
}

func usage() {
	fmt.Fprintf(os.Stderr, "Usage: %s\n\nGlobal options:\n", cmdr.UsageLine)
	flag.PrintDefaults()
	fmt.Fprintf(os.Stderr, "\nType \"radioctl help\" for more documentation.\n\n")
}

func main() {
	flag.Parse()
	log.SetFlags(0)
	if err := cmdr.Dispatch(flag.Args()); err != nil {
		log.Fatal(err)
	}
}
